/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
package com.openexchange.office.realtime.doc;

import com.openexchange.exception.OXException;
import com.openexchange.log.Log;
import com.openexchange.office.hazelcast.doc.DocSaveStateDirectory;
import com.openexchange.office.hazelcast.doc.DocumentDirectory;
import com.openexchange.office.tools.directory.DocResourceID;
import com.openexchange.office.tools.directory.DocSaveState;
import com.openexchange.office.tools.directory.DocumentState;

/**
 * Helper class for documents state which must be available
 * to all cluster-nodes (via Hazelcast).
 *
 * @author Carsten Driesner
 * @since 7.8.1
 *
 */
public class DocumentStateHelper {

    @SuppressWarnings("deprecation")
    static private final org.apache.commons.logging.Log LOG = Log.loggerFor(DocumentStateHelper.class);

    public static void updateSaveState(final DocSaveStateDirectory globalSaveStateDir, final String contextId, final String docId, boolean saveState) {
        // set global save state to prevent a second access via a new rt group dispatcher instance
        try {
            if (null != globalSaveStateDir) {
                final DocResourceID docResId = DocResourceID.createDocResourceID(contextId, docId);
                DocSaveState docState = globalSaveStateDir.get(docResId);
                if (null != docState) {
                    docState.setSaveInProgress(saveState);
                } else {
                    docState = new DocSaveState(docResId, saveState);
                }
                globalSaveStateDir.set(docResId, docState);
            } else {
                LOG.warn("RT connection: Failed to find hazelcast based document save state directory");
            }
        } catch (OXException e) {
            LOG.error("RT connection: Failed to set global save state for document context " + docId, e);
        }
    }

}
